/* implement word recognizer 
   Author:  William B. Brogden    Aug. 1990  */!!

inherit(Object, #Recognizer, #(strings   /* sorted collection */
recStruct /* final compact struct */
cells       /* temp ordered collection */
status
), 2, nil)!!

now(class(Recognizer))!!

now(Recognizer)!!

/* return array of points unpacked from struct  */
Def dump(self | ar, tmp, fl )
{ if recStruct 
  then tmp := 0 ;
    ar := new(OrderedCollection, 100 );
    loop
    while tmp < size(recStruct)
    begin
      fl := byteAt( recStruct, tmp) ;
      add(ar, point( fl, wordAt( recStruct, tmp + 1 ) ) );
      tmp := tmp + 3 ;   
    endLoop;
    ^ar ;
  else ^nil ;
  endIf;
}
!!

/* convert cells array to binary struct 
   and return the struct created   */
Def cellToStruct(self | tmp, ct, cellct, base, wrk )
{ if (cellct := size(cells)) == 0
  then ^nil;
  endIf;
  recStruct := new(Struct, cellct * 3 ) ; /* automatically 0 */
  base := 0 ;
  loop
  while base < cellct
  begin
    wrk := base * 3 ;       /* work adr in struct */
    ct := y(cells[base]) ;  /* ct of cells this level */
    tmp := x(cells[base]) ; /* status flag 0 or word# */
    putByte( recStruct, tmp, wrk ) ;  wrk := inc(wrk); 
    putWord( recStruct, ct, wrk ) ;   wrk := inc2(wrk);
    base := inc(base ) ;
    if base < cellct 
    then loop
         while  ct > 0 
     begin 
      putByte( recStruct, x(cells[base]), wrk ) ; wrk := inc(wrk);
      tmp := y(cells[base]) * 3 ;   /* index of cell -> struct offset */
      putWord( recStruct, tmp - wrk, wrk ) ; /* now rel offset */
      wrk := inc2(wrk);
      ct := dec(ct); 
      base := inc(base); /* point to next cell */
      endLoop; 
    endIf;   
  endLoop;
  ^recStruct;
}
!!

/*  recognize a word using "cells" rather than binary struct
 This is slower than the primative but is included so you
    can see how it works.
 aStr must be null terminated, ucase if struct built that way
    return nil if not recog, else Int 0 -n */
Def wdRecog(self, aStr | state,ch, tn, flag )
{ state := 0 ;
  do( aStr,
    { using(n)   /* char from string collection aStr */
      ch := asInt(n);
      if   ch == 0
      then   /* end of word, check flag */
        if  (flag := x(cells[state])) == 0
        then ^nil ; 
        else ^dec(flag); /* word# is 0 based */
        endIf;
      endIf;
      tn := y(cells[state]) ;  /* count of letters at this level */
      if   tn == 0
      then ^nil;
      else flag := 1 ;
      endIf;
      loop
      while flag    /* until nil */
      begin 
        state := inc(state);
        if   ch == x(cells[state])
        then state := y(cells[state]); flag := nil;
        else 
          if  (tn := dec(tn)) == 0
          then ^nil ;
          endIf;
        endIf;
      endLoop; 
    });
  ^nil ;   
}
!!

/*  the workhorse routine - called recursively 
    sCell and eCell are start and end(+1) cells in cell array
    sWord and eWord are indexes of strings
*/
Def rc2Build(self, sCell, eCell, sWord, eWord, depth |
      tch, tcell, tword, tmp )
{ loop
  while sCell < eCell
  begin
    tcell := size(cells);   /* where next layer will go */
    tmp := cells[ sCell ]; 
    put( cells, point( x(tmp), size(cells)) , sCell );
     /* set pointer to cell we are about to build */
    tword := sWord ;
    tch := asInt( at( strings[tword], depth )) ;
    if   tch == 0 
    then
      if  (tword + 1 ) < eWord
      then 
        tword := sWord := inc(sWord);  /* that word terminated, try next */
        tch := asInt( at( strings[tword], depth )) ;
      else break();
      endIf;  
    endIf;
    loop
    while (tword + 1 ) < eWord cand
          (tch == asInt( at( strings[tword + 1], depth ) ))
    begin tword := inc(tword) ;
    endLoop; /* at this point, sWord - tword range identical at depth */
       /* so we build cells for depth+1 if any */
    tmp := rc3Build(self, sWord, tword + 1, depth + 1 ) ;
       /* note we must use tword +1 as end */
    if    tmp > 0
    then tcell := inc(tcell);  /* past ct cell just built */
         rc2Build( self, tcell, size(cells), sWord, tword+1, depth + 1 ) ;
    endIf;
    sWord := tword + 1 ;
    sCell := inc(sCell);
  endLoop;
  ^nil;
}
!!

/* construct cells needed for the number of unique letters
   in the start-end range at depth  */
Def rc3Build(self, start, end, depth | tct, ch, tch, tptr )
{ tct := unique(self, start, end,depth );
  if tct > 0
  then   /* some word(s) left, first check for a sep */
    if   asInt(at( strings[start], depth )) == 0
    then push(cells, point(start+1, tct) ) ;  /* show as legal word */
    else push(cells, point( 0, tct ) )   ;  /* no word here */
    endIf;
    ch := asChar(0);
    do( over(start,end),
    { using(n)
      tch := at(strings[n],depth);
      if   ch == tch
      else push(cells, point(asInt(tch),0)); /* create new letter cell */
           ch := tch ;
      endIf;
    });
  else  /* must be at end of word, flag it! */
    push(cells, point( start+1, 0 ) ) ;
  endIf;
  ^tct ;
}
!!

/* base method for creation of recog array struct
   return count of cells in structure 
   start and end are indexes of the strings array (0 based) */
Def recogBuild(self|  start, end, tmp )
{ if  (end := size(strings) ) = 0
  then errorBox("recogBuild","no data"); ^0 ;
  endIf;
  cells := new(OrderedCollection,100 ); 
  start :=  0 ;
  if   rc3Build(self,start,end,0 ) == 0
  then errorBox("recogBuild","bad start"); ^0;
  endIf;
  rc2Build(self, 1, size(cells), start, end, 0 ) ;
  ^status := size( cells );
}
!!

/* given start and end word indexes & depth
   start = index of word, end = 1 past last word to look at
   return number of unique letters at this depth, may be 0 */
Def unique(self,start,end,depth | tmp, ch, tch )
{ tmp := 0 ;
  if  start >= size( strings )
  then errorBox("unique","out of range"); ^tmp ;
  endIf;
  if  (start + 1)  == end
  then 
    if   asInt( at( strings[start], depth )) == 0 
    then ^0 ;
    else ^1 ;
    endIf;
  else
    ch := asChar(0);
    do( over(start, end ),
    { using(n)
      tch := at( strings[n], depth ) ;
      if   ch == tch
      else tmp := inc(tmp); ch := tch ;
      endIf;
    });
    ^tmp ;
  endIf;
}
!!

/* add a string - presumably a single word - 
   to the  sorted collection of words to recognize
  - return current size of coll */
Def add(self, aStr )
{ add(strings, asciiz(aStr) );
  ^size(strings);
}
!!

/* return variable */
Def recStruct(self)
{ ^recStruct ;
}
!!

/* return variable */
Def status(self)
{ ^status ;
}
!!

/* return variable */
Def strings(self)
{ ^strings ;
}
!!

/* initialize new recognizer */
Def init(self)
{ strings := new(SortedCollection,10);
  status := 0 ;
}
!!